#include <math.h>
#include "buildbsp.h"
#include "sysutils.h"
#include "map.h"
#include "geometry.h"
#include "dialog.h"
#include "sysmain.h"
#include "resource.h"

typedef struct
{
    int         numverts;
    point_t*    verts;
} polygon_t;

extern float        bsplinecoords[];
extern int          numbsplines;
extern polygon_t    subsectors[];
extern int          numsubsectors;

void CheckSubsector(Storage<line_t>* lines)
{
    int         i, j, c, s;
    line_t*     line;
    float       dx, dy, len;
    point_t     v, startv;
    divline_t   divline;
    point_t     vertices[256];

    c = lines->count();
    if (!c)
    {
        Debug("CheckSubsector: a subsector has not any seg\n");
        return;
    }
    if (c < 3)
        Debug("CheckSubsector: a subsector has only one or two segs\n");

    for (i=0 ; i<c ; i++)
    {
        line = lines->elementAt(i);
        if (line->side == -1)
        {
            bsplinecoords[numbsplines*4  ] = line->p1.x;
            bsplinecoords[numbsplines*4+1] = line->p1.y;
            bsplinecoords[numbsplines*4+2] = line->p2.x;
            bsplinecoords[numbsplines*4+3] = line->p2.y;
            numbsplines++;
        }
    }

    for (i=0 ; i<c ; i++)
    {
        line = lines->elementAt(i);
        dx = line->p2.x - line->p1.x;
        dy = line->p2.y - line->p1.y;
        len = (float)sqrt(dx*dx+dy*dy);
        if (len < 1.0f)
            Debug("CheckSubsector: a subsector has a seg of the length less than 1.0: %f\n", len);
    }

    for (i=0 ; i<c ; i++)
    {
        line = lines->elementAt(i);
        DivlineFromWorldline(&divline, line);
        for (j=0 ; j<c ; j++)
        {
            if (j == i)
                continue;
            line = lines->elementAt(j);
            s = LineOnSide(line, &divline);
            if (s)
            {
                Debug("CheckSubsector: a subsector is not convex\n");
                goto addsspoly;
            }
        }
    }

addsspoly:
    j = 0;
    line = lines->elementAt(0);
    startv = line->p1;
    vertices[j++] = startv;
    v = line->p2;
    while (!ClosePoints(&v, &startv))
    {
        vertices[j++] = v;
        for (i=1 ; i<c ; i++)
        {
            line = lines->elementAt(i);
            if (ClosePoints(&v, &line->p1))
            {
                v = line->p2;
                break;
            }
        }
        if (i == c)
        {
            Debug("CheckSubsector: a subsector is not closed\n");
            return;
        }
    }
    if (j != c)
        Debug("CheckSubsector: the numbers of the segs and the vertices do not match\n");
    subsectors[numsubsectors].numverts = j;
    if (subsectors[numsubsectors].verts)
        free(subsectors[numsubsectors].verts);
    subsectors[numsubsectors].verts = (point_t*)malloc(sizeof(point_t)*j);
    memcpy(subsectors[numsubsectors].verts, vertices, sizeof(point_t)*j);
    numsubsectors++;
}

float   dp1, dp2;
float   len;

BOOL DoesOverlap(line_t* line1, line_t* line2)
{
    float   high, low;
    float   dx1, dy1, dx2, dy2/*, len*/;
    float   cp1, cp2, dp/*, dp1, dp2*/;

    // AABB intersection test
    high = line1->p1.y > line1->p2.y ? line1->p1.y : line1->p2.y;
    low  = line2->p1.y < line2->p2.y ? line2->p1.y : line2->p2.y;
    if ((high+0.1) < (low-0.1))
        return FALSE;
    high = line2->p1.y > line2->p2.y ? line2->p1.y : line2->p2.y;
    low  = line1->p1.y < line1->p2.y ? line1->p1.y : line1->p2.y;
    if ((high+0.1) < (low-0.1))
        return FALSE;
    high = line1->p1.x > line1->p2.x ? line1->p1.x : line1->p2.x;
    low  = line2->p1.x < line2->p2.x ? line2->p1.x : line2->p2.x;
    if ((high+0.1) < (low-0.1))
        return FALSE;
    high = line2->p1.x > line2->p2.x ? line2->p1.x : line2->p2.x;
    low  = line1->p1.x < line1->p2.x ? line1->p1.x : line1->p2.x;
    if ((high+0.1) < (low-0.1))
        return FALSE;

    dx1 = line1->p2.x - line1->p1.x;
    dy1 = line1->p2.y - line1->p1.y;
    dx2 = line2->p2.x - line2->p1.x;
    dy2 = line2->p2.y - line2->p1.y;
    len = myNormalize(&dx1, &dy1);

    cp1 = OuterProduct(dx1, dy1, (line2->p1.x - line1->p1.x), (line2->p1.y - line1->p1.y));
    cp2 = OuterProduct(dx1, dy1, (line2->p2.x - line1->p1.x), (line2->p2.y - line1->p1.y));
    if ((cp1 < -0.1 || cp1 > 0.1) || (cp2 < -0.1 || cp2 > 0.1))
        return FALSE;

    dp = InnerProduct(dx1, dy1, dx2, dy2);
    if (dp > 0)
        return FALSE;

    if ((line1->p1.x == line2->p2.x && line1->p1.y == line2->p2.y) || (line1->p2.x == line2->p1.x && line1->p2.y == line2->p1.y))
        return TRUE;

    dp1 = InnerProduct(dx1, dy1, (line2->p1.x - line1->p1.x), (line2->p1.y - line1->p1.y));
    dp2 = InnerProduct(dx1, dy1, (line2->p2.x - line1->p1.x), (line2->p2.y - line1->p1.y));
    if ((dp1 > (len-0.1) && dp2 > (len-0.1)) || (dp1 < 0.1 && dp2 < 0.1))
        return FALSE;

    return TRUE;
}

typedef struct
{
    point_t lvert, rvert;
    int     adjacentcell;
} portal_t;

void AddPortal(Storage<portal_t>* portalstore, line_t* line1, line_t* line2, int cellnum)
{
    portal_t    portal;

    if ((line1->p1.x == line2->p2.x && line1->p1.y == line2->p2.y) && (line1->p2.x == line2->p1.x && line1->p2.y == line2->p1.y))
    {
        portal.lvert = line1->p1;
        portal.rvert = line1->p2;
    }
    else if (line1->p1.x == line2->p2.x && line1->p1.y == line2->p2.y)
    {
        portal.lvert = line1->p1;
        if (dp1 > len)
            portal.rvert = line1->p2;
        else
            portal.rvert = line2->p1;
    }
    else if (line1->p2.x == line2->p1.x && line1->p2.y == line2->p1.y)
    {
        portal.rvert = line1->p2;
        if (dp2 < 0)
            portal.lvert = line1->p1;
        else
            portal.lvert = line2->p2;
    }
    else if (dp1 > 0 && dp1 < len)
    {
        portal.rvert = line2->p1;
        if (dp2 < 0)
            portal.lvert = line1->p1;
        else
            portal.lvert = line2->p2;
    }
    else if (dp2 > 0 && dp2 < len)
    {
        portal.lvert = line2->p2;
        if (dp1 > len)
            portal.rvert = line1->p2;
        else
            portal.rvert = line2->p1;
    }
    else
    {
        portal.lvert = line1->p1;
        portal.rvert = line1->p2;
    }

    portal.adjacentcell = cellnum;
    portalstore->addElement(&portal);
}

BOOL IsSeparable(point_t* lverts, point_t* rverts, int num)
{
    int         i, j, k, l;
    vertex_t    l1v1, l1v2, l2v1, l2v2;

    for (i=0 ; i<num ; i++)
    {
        l1v1.x = lverts[i].x;
        l1v1.y = lverts[i].y;
        for (j=i+1 ; j<num ; j++)
        {
            l1v2.x = lverts[j].x;
            l1v2.y = lverts[j].y;
            for (k=0 ; k<num ; k++)
            {
                l2v1.x = rverts[k].x;
                l2v1.y = rverts[k].y;
                for (l=k+1 ; l<num ; l++)
                {
                    l2v2.x = rverts[l].x;
                    l2v2.y = rverts[l].y;
                    if (DoesIntersect(l1v1, l1v2, l2v1, l2v2))
                        return FALSE;
                }
            }
        }
    }

    return TRUE;
}

extern Storage<Storage<line_t>*>* subsecportals;
extern byte* subsecmatrix;
Storage<Storage<portal_t>*>* cellportals;

int*        cells;
point_t*    lverts;
point_t*    rverts;

void TraverseAdjacentCells(int cellnum, int depth)
{
    Storage<portal_t>*  portals = *(cellportals->elementAt(cellnum));
    int                 numportals = portals->count();
    int                 i, j, root;

    for (i=0 ; i<numportals ; i++)
    {
        portal_t*   portal = portals->elementAt(i);
        int         nextcell = portal->adjacentcell;
        for (j=0 ; j<depth ; j++)
            if (cells[j] == nextcell)
                break;
        if (j < depth)
            continue;
        cells = (int*)realloc(cells, sizeof(int)*(depth+1));
        lverts = (point_t*)realloc(lverts, sizeof(point_t)*(depth+1));
        rverts = (point_t*)realloc(rverts, sizeof(point_t)*(depth+1));
        cells[depth] = cellnum;
        lverts[depth] = portal->lvert;
        rverts[depth] = portal->rvert;
        if (IsSeparable(lverts, rverts, depth+1))
            TraverseAdjacentCells(nextcell, depth+1);
    }

    root = cells[0];
    subsecmatrix[root*numsubsectors+cellnum] = 1;
}

void BuildPVS(void)
{
    int             c, i, j, k, l, linecount, linecount2;
    Storage<line_t> *lines, *lines2;
    line_t          *line, *line2;
    char            buf[256];

    PrintDialog("PVSvZ...", hProgDialog, IDC_STEXT1);

    cellportals = new Storage<Storage<portal_t>*>;

    c = subsecportals->count();
    if (c != numsubsectors)
        Error("BuildPVS: numsubsectors  subsecportals ̗vf܂\n");

    Debug("numsubsectors = %d\n", c);

    for (i=0 ; i<c ; i++)
    {
        Storage<portal_t>* portalstore = new Storage<portal_t>;
        lines = *(subsecportals->elementAt(i));
        linecount = lines->count();
        for (j=0, line = lines->elementAt(0) ; j<linecount ; j++, line++)
        {
            for (k=0 ; k<c ; k++)
            {
                if (k == i)
                    continue;
                lines2 = *(subsecportals->elementAt(k));
                linecount2 = lines2->count();
                for (l=0, line2 = lines2->elementAt(0) ; l<linecount2 ; l++, line2++)
                {
                    if (DoesOverlap(line, line2))
                        AddPortal(portalstore, line, line2, k);
                }
            }
        }
        cellportals->addElement(&portalstore);
    }

    if (subsecmatrix)
        free(subsecmatrix);
    subsecmatrix = (byte*)calloc(c*c, 1);

    cells = (int*)malloc(sizeof(int));
    lverts = (point_t*)malloc(sizeof(point_t));
    rverts = (point_t*)malloc(sizeof(point_t));

    for (i=0 ; i<c ; i++)
    {
        wsprintf(buf, "PVSvZ... %6d / %d", i, c);
        PrintDialog(buf, hProgDialog, IDC_STEXT1);

        Storage<portal_t>*  portals = *(cellportals->elementAt(i));
        int                 numportals = portals->count();
        for (j=0 ; j<numportals ; j++)
        {
            portal_t*   portal = portals->elementAt(j);
            int         nextcell = portal->adjacentcell;
            cells = (int*)realloc(cells, sizeof(int));
            lverts = (point_t*)realloc(lverts, sizeof(point_t));
            rverts = (point_t*)realloc(rverts, sizeof(point_t));
            cells[0] = i;
            lverts[0] = portal->lvert;
            rverts[0] = portal->rvert;
            TraverseAdjacentCells(nextcell, 1);
        }
    }

    free(cells);
    free(lverts);
    free(rverts);

    for (i=0 ; i<c ; i++)
        delete *(cellportals->elementAt(i));
    delete cellportals;
}
