#include <windows.h>
#include <math.h>
#include "map.h"

float myLength(float x, float y)
{
    return (float)sqrt(x*x+y*y);
}

float myNormalize(float* dx, float* dy)
{
    float   len;

    len = myLength(*dx, *dy);
    *dx = *dx / len;
    *dy = *dy / len;
    return len;
}

float InnerProduct(float x1, float y1, float x2, float y2)
{
    return x1*x2 + y1*y2;
}

float OuterProduct(float x1, float y1, float x2, float y2)
{
    return x1*y2 - x2*y1;
}

BOOL Sign(float val)
{
    if (val >= 0)
        return TRUE;

    return FALSE;
}

BOOL DoesIntersect(vertex_t l1v1, vertex_t l1v2, vertex_t l2v1, vertex_t l2v2)
{
    float   dx, dy, px, py;
    BOOL    s1, s2;

    dx = l1v2.x - l1v1.x;
    dy = l1v2.y - l1v1.y;
    px = l2v1.x - l1v1.x;
    py = l2v1.y - l1v1.y;
    s1 = Sign(OuterProduct(dx, dy, px, py));
    px = l2v2.x - l1v1.x;
    py = l2v2.y - l1v1.y;
    s2 = Sign(OuterProduct(dx, dy, px, py));

    if (s1 != s2)
    {
        dx = l2v2.x - l2v1.x;
        dy = l2v2.y - l2v1.y;
        px = l1v1.x - l2v1.x;
        py = l1v1.y - l2v1.y;
        s1 = Sign(OuterProduct(dx, dy, px, py));
        px = l1v2.x - l2v1.x;
        py = l1v2.y - l2v1.y;
        s2 = Sign(OuterProduct(dx, dy, px, py));

        if (s1 != s2)
            return TRUE;
    }

    return FALSE;
}

void SnapToLine(vertex_t* v, linedef_t* line)
{
    float   dx, dy, px, py;
    float   scale;

    dx = line->v2->x - line->v1->x;
    dy = line->v2->y - line->v1->y;
    myNormalize(&dx, &dy);

    px = v->x - line->v1->x;
    py = v->y - line->v1->y;

    scale = InnerProduct(dx, dy, px, py);

    v->x = line->v1->x + dx*scale;
    v->y = line->v1->y + dy*scale;
}

BOOL IsOnLine(vertex_t v, linedef_t line)
{
    float   dx, dy, px, py;
    float   len;
    float   val;

    dx = line.v2->x - line.v1->x;
    dy = line.v2->y - line.v1->y;
    px = v.x - line.v1->x;
    py = v.y - line.v1->y;

    len = myNormalize(&dx, &dy);

    val = OuterProduct(dx, dy, px, py);
    if (val > -0.1f && val < 0.1f)
    {
        val = InnerProduct(dx, dy, px, py);
        if (val > -0.1f && val < (len+0.1f))
            return TRUE;
    }

    return FALSE;
}

// Real-Time Rendering SE p585 Q
BOOL IsInSector(vertex_t v, sector_t* sec)
{
    linedef_t*  line;
    vertex_t    *v1, *v2;
    float       v1x, v1y, v2x, v2y;
    BOOL        s1, s2, s3;
    BOOL        ret;

    ret = FALSE;
    for (line = firstlinedef.next ; line ; line = line->next)
    {
        if (line->sides[0].sec == sec || line->sides[1].sec == sec)
        {
            v1 = line->v1;
            v2 = line->v2;
            s1 = Sign(v1->y - v.y);
            s2 = Sign(v2->y - v.y);
            if (s1 != s2)
            {
                v1x = v1->x - v.x;
                v1y = v1->y - v.y;
                v2x = v2->x - v.x;
                v2y = v2->y - v.y;
                s3 = Sign(OuterProduct(v1x, v1y, v2x, v2y));
                if (s2 == s3)
                    ret = !ret;
            }
        }
    }

    return ret;
}

BOOL IsCCW(vertex_t* verts, int numverts)
{
    int     i;
    float   dx, dy, temp;
    float   ox, oy;
    float   v1x, v1y, v2x, v2y;
    BOOL    s1, s2, s3;
    BOOL    ret;

    dx = verts[0].x - verts[numverts-1].x;
    dy = verts[0].y - verts[numverts-1].y;
    ox = verts[numverts-1].x + 0.5f*dx;
    oy = verts[numverts-1].y + 0.5f*dy;
    temp = dx;
    dx = -dy;   // is
    dy = temp;  // @xNg

    ret = FALSE;
    v1x = verts[0].x - ox;
    v1y = verts[0].y - oy;
    s1 = Sign(OuterProduct(dx, dy, v1x, v1y));
    for (i=1 ; i<numverts ; i++)
    {
        v2x = verts[i].x - ox;
        v2y = verts[i].y - oy;
        s2 = Sign(OuterProduct(dx, dy, v2x, v2y));
        if (s1 != s2)
        {
            s3 = Sign(OuterProduct(v1x, v1y, v2x, v2y));
            if (s2 == s3)
                ret = !ret;
        }
        s1 = s2;
        v1x = v2x;
        v1y = v2y;
    }

    return ret;
}