typedef enum
{
    ot_none, // = 0
    ot_vertex,
    ot_line,
    ot_sector,
    ot_thing

} objecttype_t;

typedef struct
{
    objecttype_t    type;
    void*           ptr;

} object_t;

typedef struct vertex_s
{
    float   x;
    float   y;

    object_t    snapped;

    struct vertex_s*    next;

} vertex_t;

typedef struct sector_s
{
    int     floorheight;
    int     ceilingheight;

    char    floorflat[8];
    char    ceilingflat[8];
    
    int     lightlevel;
    int     special;
    int     tag;

    struct sector_s*    next;

} sector_t;

typedef struct
{
    int     xoffset;
    int     yoffset;

    char    upper[8];
    char    lower[8];
    char    middle[8];

    sector_t*   sec;

} sidedef_t;

typedef struct linedef_s
{
    vertex_t*   v1;
    vertex_t*   v2;

    int         flags;
    int         special;
    int         tag;

    sidedef_t   sides[2];

    struct linedef_s*   next;

} linedef_t;

typedef struct thing_s
{
    float   x;
    float   y;
    int     angle;
    int     type;
    int     options;

    int     id;

    struct thing_s* next;

} thing_t;

void ClearMap(void);

void* LinePtrForNum(unsigned int num);
void* VertexPtrForNum(unsigned int num);
void* ThingPtrForNum(unsigned int num);
int VertexNumForPtr(vertex_t* ptr);
int SectorNumForPtr(sector_t* ptr);
void* SectorPtrForNum(unsigned int num);

void ScreenToWorld(int x, int y, vertex_t* v);
void InsertVertex(vertex_t v, linedef_t* line);
sector_t* PointInSector(vertex_t v);
void AddThing(int id, int x, int y);
BOOL CheckVertex(vertex_t* v);
void CalculateOffset(float centerx, float centery, int x, int y);
void MoveVertex(vertex_t* vert, int x, int y);
void MoveThing(thing_t* thing, int x, int y);
void DeleteVertex(vertex_t* v);
void DeleteThing(thing_t* thing);
void DeleteSector(sector_t* sec);

void LS_AddVertex(int x, int y);
void LS_MoveVertex(int x, int y);
void LS_Abort(void);
void LS_Close(void);
void LS_DecrementVertex(void);

extern int numsectors;
extern int numlinedefs;
extern int numvertices;
extern int numthings;

extern sector_t firstsector;    
extern linedef_t firstlinedef;
extern vertex_t firstvertex;
extern thing_t firstthing;

extern object_t selected;
extern object_t picked;

extern char defmidtexture[];
extern char defupptexture[];
extern char deflowtexture[];

extern sector_t defsecstate;

extern BOOL mapalterd;

extern BOOL     secdrawmode;
extern vertex_t lsverts[];
extern int      numlsverts;