#include <windows.h>
#include <gl\gl.h>             
#include <gl\glu.h>
#include "sysmain.h"
#include "map.h"
#include "geometry.h"
#include "thingdef.h"
#include "resource.h"

HDC     hDC;
HGLRC   hRC = NULL;
float   viewx, viewy, viewwidth, viewheight;
float   viewscale = 1.0f;
int     gridsize;
BOOL    drawaxis;
BITMAPINFO bi = {
    {40, 88, 88, 1, 24, BI_RGB}
};
byte    bits[88][88][3];
HBITMAP hBitmap;
byte    data[88][88][4];
GLuint  tgtex;
int     translate[9] = { 48, 37, 55, 33, 53, 43, 36, 54, 35 };

void SetupGL(HWND hwnd)
{
    PIXELFORMATDESCRIPTOR pfd = {
	sizeof(PIXELFORMATDESCRIPTOR),  // Size of this structure
	1,                              // Version of this structure    
	PFD_DRAW_TO_WINDOW |            // Draw to Window (not to bitmap)
	PFD_SUPPORT_OPENGL |            // Support OpenGL calls in window
	PFD_DOUBLEBUFFER,               // Double buffered
	PFD_TYPE_RGBA,                  // RGBA Color mode
	24,                             // Want 24bit color 
	0,0,0,0,0,0,                    // Not used to select mode
	0,0,                            // Not used to select mode
	0,0,0,0,0,                      // Not used to select mode
	32,                             // Size of depth buffer
	0,                              // Not used to select mode
	0,                              // Not used to select mode
	0, //PFD_MAIN_PLANE,            // Draw in main plane
	0,                              // Not used to select mode
	0,0,0 };                        // Not used to select mode
	int nPixelFormat;
    RECT rect;

    hDC = GetDC(hwnd);
	nPixelFormat = ChoosePixelFormat(hDC, &pfd);
	SetPixelFormat(hDC, nPixelFormat, &pfd);
    hRC = wglCreateContext(hDC);
    wglMakeCurrent(hDC, hRC);
    glEnable(GL_TEXTURE_2D);

    GetClientRect(hwnd, &rect);
    viewx = 0;
    viewy = 0;
    viewwidth = rect.right*1.0f;
    viewheight = viewwidth*rect.bottom/rect.right;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(viewx-(viewwidth/2), viewx+(viewwidth/2), viewy-(viewheight/2), viewy+(viewheight/2));
    glClearColor(0.5, 0.5, 0.5, 1.0);

    hBitmap = LoadBitmap(hInst, MAKEINTRESOURCE(IDB_GAMEFONT2));
    GetDIBits(hDC, hBitmap, 0, 88, bits, &bi, DIB_RGB_COLORS);
    for (int i=0 ; i<88 ; i++)
    {
        for (int j=0 ; j<88 ; j++)
        {
            data[87-i][j][2] = bits[i][j][0];
            data[87-i][j][1] = bits[i][j][1];
            data[87-i][j][0] = bits[i][j][2];
            if (bits[i][j][2] == 126 &&
                bits[i][j][1] == 127 &&
                bits[i][j][0] == 128)
                data[87-i][j][3] = 0;
            else
                data[87-i][j][3] = 255;
        }
    }
    glGenTextures(1, &tgtex);
	glBindTexture(GL_TEXTURE_2D, tgtex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
    glTexImage2D(GL_TEXTURE_2D, 0, 4, 88, 88, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
    glBindTexture(GL_TEXTURE_2D, 0);
}

void CleanupGL(void)
{
    if (hRC)
    {
        wglMakeCurrent(hDC, NULL);
        wglDeleteContext(hRC);
    }
}

void DrawGrid(void)
{
    int x, y;

    if (!gridsize)
        return;

    glColor3ub(112, 112, 112);
    glBegin(GL_LINES);
    for (y = (int)(viewy+viewheight/2+gridsize)/gridsize*gridsize ; y >= (int)(viewy-viewheight/2-gridsize) ; y -= gridsize)
    {
        glVertex2i((int)(viewx-viewwidth/2-5), y);
        glVertex2i((int)(viewx+viewwidth/2+5), y);
    }
    for (x = (int)(viewx-viewwidth/2-gridsize)/gridsize*gridsize ; x <= (int)(viewx+viewwidth/2+gridsize) ; x += gridsize)
    {
        glVertex2i(x, (int)(viewy+viewheight/2+5));
        glVertex2i(x, (int)(viewy-viewheight/2-5));
    }
    glEnd();
}

void DrawAxes(void)
{
    if (!drawaxis)
        return;

    glColor3ub(0, 0, 0);
    glBegin(GL_LINES);
    glVertex2f(viewx-viewwidth/2-5, 0);
    glVertex2f(viewx+viewwidth/2+5, 0);
    glVertex2f(0, viewy+viewheight/2+5);
    glVertex2f(0, viewy-viewheight/2-5);
    glEnd();
}

void DrawLines(void)
{
    linedef_t*  line;
    int         i=0;

    for (line = firstlinedef.next ; line ; line = line->next, i++)
    {
        glLoadName(i); // glBegin()-glEnd()Ԃł͎gȂ
        if (((secdrawmode || insvertmode) && (picked.type == ot_line && picked.ptr == line)) ||
            (selected.type == ot_line && selected.ptr == line) ||
            (selected.type == ot_sector && (line->sides[0].sec == selected.ptr || line->sides[1].sec == selected.ptr)))
            glColor3ub(255, 0, 0);
        else
            glColor3ub(255, 255, 255);
        glBegin(GL_LINES);
        glVertex2fv((float *)line->v1);
        glVertex2fv((float *)line->v2);
        if (selected.type == ot_line && selected.ptr == line)
        {
            float   dx, dy, x;
            float   v1x, v1y, v2x, v2y;

            dx = line->v2->x - line->v1->x;
            dy = line->v2->y - line->v1->y;
            v1x = line->v1->x + dx*0.5f;
            v1y = line->v1->y + dy*0.5f;
            x = dx;
            dx = dy;
            dy = -x;
            myNormalize(&dx, &dy);
            v2x = v1x + dx*8/viewscale;
            v2y = v1y + dy*8/viewscale;

            glVertex2f(v1x, v1y);
            glVertex2f(v2x, v2y);
        }
        glEnd();
    }
}

#define POINTSIZE   5

void DrawVertices(void)
{
    vertex_t*   v;
    int         i=10000;
    linedef_t*  line;

    glPointSize(POINTSIZE);

    for (v = firstvertex.next ; v ; v = v->next, i++)
    {
        glLoadName(i);
        if ((secdrawmode && (picked.type == ot_vertex && picked.ptr == v)) ||
            (selected.type == ot_vertex && selected.ptr == v))
            glColor3ub(255, 0, 0);
        else if (selected.type == ot_sector)
        {
            for (line = firstlinedef.next ; line ; line = line->next)
            {
                if (line->sides[0].sec == selected.ptr || line->sides[1].sec == selected.ptr)
                {
                    if (line->v1 == v || line->v2 == v)
                    {
                        glColor3ub(255, 0, 0);
                        break;
                    }
                }
            }
            if (!line)
                glColor3ub(255, 255, 255);
        }
        else
            glColor3ub(255, 255, 255);
        glBegin(GL_POINTS);
        glVertex2fv((float *)v);
        glEnd();
    }

    if (secdrawmode && numlsverts)
    {
        glLoadName(19999);
        if (picked.type == ot_vertex && picked.ptr == 0)
            glColor3ub(255, 0, 0);
        else
            glColor3ub(255, 255, 0);
        glBegin(GL_POINTS);
        glVertex2fv((float *)&lsverts[0]);
        glEnd();
    }
}

void DrawThings(void)
{
    thing_t*    thing;
    int         i=20000;
    int         r;

    for (thing = firstthing.next ; thing ; thing = thing->next, i++)
    {
        glLoadName(i);
        if (selected.type == ot_thing && selected.ptr == thing)
            glColor3ub(255, 0, 0);
        else
            glColor3ub(255, 255, 255);
        r = thingdefs[thing->id].radius;
        if (!r)
        {
            glBegin(GL_LINES);
            glVertex2f(thing->x - 8, thing->y - 8);
            glVertex2f(thing->x + 8, thing->y + 8);
            glVertex2f(thing->x + 8, thing->y - 8);
            glVertex2f(thing->x - 8, thing->y + 8);
            glEnd();
        }
        else
        {
            int group = thingdefs[thing->id].group;
            int index = translate[group];
            float s1 = (index%8)/8.0f;
            float s2 = s1 + 0.125f;
            float t1 = (index/8)/8.0f;
            float t2 = t1 + 0.125f;
            if (group == tg_ammunition || group == tg_powerup || group == tg_victim)
            {
                s1 += 0.006f;
                s2 += 0.006f;
            }
            glBindTexture(GL_TEXTURE_2D, tgtex);
            glBegin(GL_QUADS);
            glTexCoord2f(s1, t2);
            glVertex2f(thing->x - r, thing->y - r);
            glTexCoord2f(s2, t2);
            glVertex2f(thing->x + r, thing->y - r);
            glTexCoord2f(s2, t1);
            glVertex2f(thing->x + r, thing->y + r);
            glTexCoord2f(s1, t1);
            glVertex2f(thing->x - r, thing->y + r);
            glEnd();
            glBindTexture(GL_TEXTURE_2D, 0);
        }
    }
}

#define MAXNUMBSPLINES  4096
float   bsplinecoords[MAXNUMBSPLINES*4];
int     numbsplines;
BOOL    drawbspline;

void DrawBSPLines(void)
{
    int i;

    if (!drawbspline)
        return;

    glColor3ub(160, 160, 160);
    //glLineStipple(4, 0x9999);
    //glEnable(GL_LINE_STIPPLE);
    glBegin(GL_LINES);
    for (i=0 ; i<numbsplines ; i++)
    {
        glVertex2f(bsplinecoords[i*4  ], bsplinecoords[i*4+1]);
        glVertex2f(bsplinecoords[i*4+2], bsplinecoords[i*4+3]);
    }
    glEnd();
    //glDisable(GL_LINE_STIPPLE);
}

typedef struct
{
    float   x, y;
} point_t;

typedef struct
{
    int         numverts;
    point_t*    verts;
} polygon_t;

#define MAXNUMSUBSECTORS    4096
polygon_t   subsectors[MAXNUMSUBSECTORS];
int         numsubsectors;
byte*       subsecmatrix;
BOOL        pvstestmode;
int         pickedss = -1;

void DrawSubsectors(void)
{
    int i, j;

    if (!pvstestmode)
        return;

    for (i=0 ; i<numsubsectors ; i++)
    {
        glLoadName(i+30000);
        if (pickedss == -1)
            glColor3ub(128, 128, 128);
        else if (i == pickedss)
            glColor3ub(255, 0, 0);
        else if (subsecmatrix[numsubsectors*pickedss+i])
            glColor3ub(255, 128, 128);
        else
            glColor3ub(128, 128, 128);
        glBegin(GL_POLYGON);
        for (j=0 ; j<subsectors[i].numverts ; j++)
            glVertex2fv((float*)&subsectors[i].verts[j]);
        glEnd();
    }
}

void DrawLineStrip(void)
{
    int i;

    if (!secdrawmode)
        return;

    glColor3ub(255, 255, 0);
    glBegin(GL_LINE_STRIP);
    for (i=0 ; i<=numlsverts ; i++)
        glVertex2fv((float *)&lsverts[i]);
    glEnd();
}

void RenderScene(void)
{
    if (!hRC)
        return;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    DrawSubsectors();
    DrawGrid();
    DrawAxes();
    DrawBSPLines();
    DrawLines();
    DrawVertices();
    DrawThings();
    DrawLineStrip();
    SwapBuffers(hDC);
}

void AdjustView(int width, int height)
{
    if (!hRC)
        return;

    glViewport(0, 0, width, height);
    viewwidth = width/viewscale;
    viewheight = viewwidth*height/width;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(viewx-(viewwidth/2), viewx+(viewwidth/2), viewy-(viewheight/2), viewy+(viewheight/2));
}

void TranslateView(int x, int y)
{
    viewx += x/viewscale;
    viewy += y/viewscale;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(viewx-(viewwidth/2), viewx+(viewwidth/2), viewy-(viewheight/2), viewy+(viewheight/2));
}

void CenterView(void)
{
    viewx = viewy = 0;
    TranslateView(0, 0);
}

void ScaleView(float factor)
{
    viewscale *= factor;
    viewwidth = screenwidth/viewscale;
    viewheight = screenheight/viewscale;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(viewx-(viewwidth/2), viewx+(viewwidth/2), viewy-(viewheight/2), viewy+(viewheight/2));
}

void ResetScale(void)
{
    viewscale = 1.0f;
    ScaleView(1.0f);
}

void FitMapToView(void)
{
    vertex_t*   v;
    float       minx, maxx, miny, maxy;
    float       w, h;
    float       scraspect, mapaspect;

    if (!numlinedefs)
        return;

    minx = miny = 32767.0f;
    maxx = maxy = -32768.0f;

    for (v = firstvertex.next ; v ; v = v->next)
    {
        if (v->x < minx)
            minx = v->x;
        if (v->x > maxx)
            maxx = v->x;
        if (v->y < miny)
            miny = v->y;
        if (v->y > maxy)
            maxy = v->y;
    }

    w = maxx - minx;
    h = maxy - miny;

    viewx = minx + w/2;
    viewy = miny + h/2;

    mapaspect = w/h;
    scraspect = viewwidth/viewheight;

    if (mapaspect > scraspect)
        viewscale = screenwidth/w;
    else
        viewscale = screenheight/h;

    // ڕς viewx, viewy, viewscale ̒l𔽉fA]T邽߃XP[
    ScaleView(0.9f);
}

void PickObjectsUnderCursor(int x, int y)
{
    GLuint  selectBuff[64];
    GLint   hits, viewport[4];
    GLuint  foremost;
    int     i;

    glSelectBuffer(64, selectBuff);
    glGetIntegerv(GL_VIEWPORT, viewport);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluPickMatrix(x, viewport[3]-y, 6, 6, viewport);
    gluOrtho2D(viewx-(viewwidth/2), viewx+(viewwidth/2), viewy-(viewheight/2), viewy+(viewheight/2));
    glRenderMode(GL_SELECT);
    glInitNames();
    glPushName(0);
    DrawLines();
    DrawThings();
    glLoadIdentity();
    gluPickMatrix(x, viewport[3]-y, 3+POINTSIZE, 3+POINTSIZE, viewport);
    gluOrtho2D(viewx-(viewwidth/2), viewx+(viewwidth/2), viewy-(viewheight/2), viewy+(viewheight/2));
    DrawVertices();
    glLoadIdentity();
    gluPickMatrix(x, viewport[3]-y, 1, 1, viewport);
    gluOrtho2D(viewx-(viewwidth/2), viewx+(viewwidth/2), viewy-(viewheight/2), viewy+(viewheight/2));
    DrawSubsectors();

    hits = glRenderMode(GL_RENDER);
    if (hits > 0)
    {
        foremost = selectBuff[3];
        for (i=1 ; i<hits ; i++)
        {
            if (selectBuff[i*4+3] > foremost)
                foremost = selectBuff[i*4+3];
        }

        if (foremost < 10000)
        {
            picked.type = ot_line;
            picked.ptr = LinePtrForNum(foremost);
        }
        else if (foremost < 20000)
        {
            picked.type = ot_vertex;
            if (foremost == 19999)
                picked.ptr = 0;
            else
                picked.ptr = VertexPtrForNum(foremost-10000);
        }
        else if (foremost < 30000)
        {
            picked.type = ot_thing;
            picked.ptr = ThingPtrForNum(foremost-20000);
        }
        else
            pickedss = foremost-30000;
    } 
    else
    {
        picked.type = ot_none;
        pickedss = -1;
    }

    glPopMatrix();
}