#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <stdio.h>
#include <math.h>
#include "doomdata.h"
#include "wad.h"
#include "buildbsp.h"
#include "map.h"
#include "sysutils.h"

FILE*                   fp;
Storage<lumpinfo_t>*    infostore;

Storage<mapthing_t>*        mapthingstore;
Storage<mapsector_t>*       mapsectorstore;
Storage<mapvertex_t>*       mapvertexstore;
Storage<mapsidedef_t>*      mapsidedefstore;
Storage<maplinedef_t>*      maplinedefstore;
Storage<mapsubsector_t>*    mapsubsecstore;
Storage<mapseg_t>*          mapsegstore;
Storage<mapnode_t>*         mapnodestore;


void addName(char* name, void* data, int size)
{
    lumpinfo_t  lump;

    memset(lump.name, 0, sizeof(lump.name));
    strncpy(lump.name, name, 8);
    for (int i=0 ; i<8 ; i++)
        lump.name[i] = toupper(lump.name[i]);
    lump.filepos = ftell(fp);
    lump.size = size;
    infostore->addElement(&lump);

    fwrite(data, 1, size, fp);
}

void OutputThings (void)
{
    int count = mapthingstore->count();
    addName("things", mapthingstore->elementAt(0), count*sizeof(mapthing_t));
}

void OutputLineDefs (void)
{
    int count = maplinedefstore->count();
    addName("linedefs", maplinedefstore->elementAt(0), count*sizeof(maplinedef_t));
}

void OutputSideDefs (void)
{
    int count = mapsidedefstore->count();
    addName("sidedefs", mapsidedefstore->elementAt(0), count*sizeof(mapsidedef_t));
}

void OutputVertexes (void)
{
    int count = mapvertexstore->count();
    addName("vertexes", mapvertexstore->elementAt(0), count*sizeof(mapvertex_t));
}

void OutputSegs (void)
{
    int count = mapsegstore->count();
    addName("segs", mapsegstore->elementAt(0), count*sizeof(mapseg_t));
}

void OutputSubsectors (void)
{
    int count = mapsubsecstore->count();
    addName("ssectors", mapsubsecstore->elementAt(0), count*sizeof(mapsubsector_t));
}

void OutputNodes (void)
{
    int count = mapnodestore->count();
    addName("nodes", mapnodestore->elementAt(0), count*sizeof(mapnode_t));
}

void OutputSectors (void)
{
    int count = mapsectorstore->count();
    addName("sectors", mapsectorstore->elementAt(0), count*sizeof(mapsector_t));
}

void OutputConnections (void)
{
    int     numsectors, bytes;
    void*   bits;

    numsectors = mapsectorstore->count();
    bytes = (numsectors*numsectors+7)/8;
    bits = calloc(bytes, 1);
    addName("reject", bits, bytes);
}

/*
===============================================================================

							PROCESSING

===============================================================================
*/

void BuildPVS(void);
extern int  numsubsectors;
extern int  calcpvs;

void ProcessConnections(void)
{
    if (calcpvs)
        BuildPVS();
    else
        numsubsectors = 0;
}

void ProcessSectors(void)
{
    sector_t*   def;
    mapsector_t ms;

    mapsectorstore = new Storage<mapsector_t>;

    for (def = firstsector.next ; def ; def = def->next)
    {
        ms.floorheight = def->floorheight;
        ms.ceilingheight = def->ceilingheight;
        memcpy (ms.floorpic,def->floorflat,8);
        memcpy (ms.ceilingpic,def->ceilingflat,8);
        ms.lightlevel = def->lightlevel;
        ms.special = def->special;
        ms.tag = def->tag;

        mapsectorstore->addElement(&ms);
    }
}

short UniqueVertex(short x, short y)
{
    int         i, count;
    mapvertex_t mv, *mvp;

    mv.x = x;
    mv.y = y;

// see if an identical vertex already exists
	count = mapvertexstore->count();
    mvp = mapvertexstore->elementAt(0);
    for (i=0 ; i<count ; i++, mvp++)
        if (mvp->x == mv.x && mvp->y == mv.y)
            return i;

    mapvertexstore->addElement(&mv);

    return count;	
}

enum {BOXTOP,BOXBOTTOM,BOXLEFT,BOXRIGHT};	// bbox coordinates

float   bbox[4];

void AddPointToBBox (point_t *pt)
{
    if (pt->x < bbox[BOXLEFT])
        bbox[BOXLEFT] = pt->x;
    if (pt->x > bbox[BOXRIGHT])
        bbox[BOXRIGHT] = pt->x;

    if (pt->y > bbox[BOXTOP])
        bbox[BOXTOP] = pt->y;
    if (pt->y < bbox[BOXBOTTOM])
        bbox[BOXBOTTOM] = pt->y;
}

#define PI  3.14159265359
Storage<Storage<line_t>*>*  subsecportals;

int/*void*/ ProcessLines (Storage<line_t>* store)
{
    int         i,count;
    line_t      *wline;
    mapseg_t    line;
    short       angle;
    double      fangle;
    linedef_t*  ld;
    int         ret=0;
    Storage<line_t>* portalstore = new Storage<line_t>;

    // 萔́A(U(u)TtBbNXtȂ)tƌȂ̂ (float)MAXINT ȂǂƏĂ\Ȃ
    bbox[BOXLEFT] = 2147483647.0f/*MAXINT*/;
    bbox[BOXRIGHT] = -2147483648.0f/*MININT*/;
    bbox[BOXTOP] = -2147483648.0f/*MININT*/;
    bbox[BOXBOTTOM] = 2147483647.0f/*MAXINT*/;

    count = store->count();
    for (i=0 ; i<count ; i++)
    {
        wline = store->elementAt(i);
        if (wline->grouped)
            Debug/*printf*/ ("ERROR: line regrouped\n");
        wline->grouped = true;

        if (wline->side == -1)
        {
            portalstore->addElement(wline);
            //AddPointToBBox (&wline->p1);
            //AddPointToBBox (&wline->p2);
            ret++;
            continue;
        }

        ld = (linedef_t*)LinePtrForNum(wline->linedef);
        if (ld->sides[1].sec)
            portalstore->addElement(wline);

        memset (&line, 0, sizeof(line));
        AddPointToBBox (&wline->p1);
        AddPointToBBox (&wline->p2);
        line.v1 = UniqueVertex ((short)wline->p1.x, (short)wline->p1.y);
        line.v2 = UniqueVertex ((short)wline->p2.x, (short)wline->p2.y);
        line.linedef = (short)wline->linedef;
        line.side = (short)wline->side;
        line.offset = (short)wline->offset;
        fangle = atan2 (wline->p2.y - wline->p1.y, wline->p2.x - wline->p1.x); // 180x360x܂ł̒l(΁`2)łǍ^ϊŃI[o[t[K؂ɏ΍\Ȃ
        angle = (short)(fangle/(PI*2)*0x10000); // 0=east, 0x4000(16384)=north, 0x8000(-32768)=west, 0xC000(-16384)=south (UDS [4-6]: SEGS Q)
        line.angle = angle;
        mapsegstore->addElement(&line);
    }

    subsecportals->addElement(&portalstore);
    return ret;
}

unsigned short ProcessSubsector (Storage<line_t>* wmaplinestore)
{
    int             count;
    mapsubsector_t  sub;
	
    memset (&sub,0,sizeof(sub));

    count = wmaplinestore->count();
    if (count < 1)
        Error ("ProcessSubsector: count = %i",count);

    sub.numsegs = count;
    sub.firstseg = mapsegstore->count();
    int numinvissegs = ProcessLines (wmaplinestore);
    sub.numsegs -= numinvissegs;

// add the new subsector
    mapsubsecstore->addElement(&sub);

    return mapsubsecstore->count() - 1;
}

void CheckSubsector(Storage<line_t>* lines);

unsigned short ProcessNode (bspnode_t *node, short *totalbox)
{
    short           subbox[2][4];
    int             i;
    unsigned short  r;
    mapnode_t       mnode;

    memset (&mnode,0,sizeof(mnode));

    if (node->lines)
    {
        CheckSubsector(node->lines); // subsectors[]ւ̗vf̒ǉs
        r = ProcessSubsector (node->lines);
        delete node->lines;
        for (i=0 ; i<4 ; i++)
            totalbox[i] = (short)bbox[i];
        return r | NF_SUBSECTOR;    // NF_SUBSECTOR flags a subsector
    }

    mnode.x = (short)node->divline.pt.x;
    mnode.y = (short)node->divline.pt.y;
    mnode.dx = (short)node->divline.dx;
    mnode.dy = (short)node->divline.dy;

    r = ProcessNode (node->side[0], subbox[0]);
    mnode.children[0] = r;
    for (i=0 ; i<4 ; i++)
        mnode.bbox[0][i] = subbox[0][i];

    r = ProcessNode (node->side[1], subbox[1]);
    mnode.children[1] = r;
    for (i=0 ; i<4 ; i++)
        mnode.bbox[1][i] = subbox[1][i];

    totalbox[BOXLEFT] = MIN(subbox[0][BOXLEFT], subbox[1][BOXLEFT]);
    totalbox[BOXTOP] = MAX(subbox[0][BOXTOP], subbox[1][BOXTOP]);
    totalbox[BOXRIGHT] = MAX(subbox[0][BOXRIGHT], subbox[1][BOXRIGHT]);
    totalbox[BOXBOTTOM] = MIN(subbox[0][BOXBOTTOM], subbox[1][BOXBOTTOM]);

    mapnodestore->addElement(&mnode);
    return mapnodestore->count() - 1;	
}

short   worldbounds[4];

void ProcessNodes(void)
{
    //short   worldbounds[4];

    mapsubsecstore = new Storage<mapsubsector_t>;
    mapsegstore = new Storage<mapseg_t>;
    mapnodestore = new Storage<mapnode_t>;

    subsecportals = new Storage<Storage<line_t>*>;

    ProcessNode (startnode, worldbounds);
}

short UniqueSidedef(sidedef_t* ws)
{
    int             i, count;
    mapsidedef_t    ms, *msp;

    ms.textureoffset = (short)ws->xoffset;
    ms.rowoffset = (short)ws->yoffset;
    memcpy (ms.toptexture, ws->upper, 8);
    memcpy (ms.bottomtexture, ws->lower, 8);
    memcpy (ms.midtexture, ws->middle, 8);
    ms.sector = (short)SectorNumForPtr(ws->sec);

// see if an identical sidedef already exists
    count = mapsidedefstore->count();
    msp = mapsidedefstore->elementAt(0);
    for (i=0 ; i<count ; i++, msp++)
        if (!memcmp(msp, &ms, sizeof(ms)))
            return i;

    mapsidedefstore->addElement(&ms);

    return count;
}

void ProcessLineSideDefs(void)
{
    maplinedef_t    ld;
    linedef_t*      wl;

    mapvertexstore = new Storage<mapvertex_t>;
    mapsidedefstore = new Storage<mapsidedef_t>;
    maplinedefstore = new Storage<maplinedef_t>;

    for (wl = firstlinedef.next ; wl ; wl = wl->next)
    {
        ld.v1 = UniqueVertex((short)wl->v1->x, (short)wl->v1->y);
        ld.v2 = UniqueVertex((short)wl->v2->x, (short)wl->v2->y);
        ld.flags = (short)wl->flags;
        ld.special = (short)wl->special;
        ld.tag = (short)wl->tag;
        ld.sidenum[0] = UniqueSidedef(&wl->sides[0]);
        if (wl->sides[1].sec)
        {
            ld.flags |= ML_TWOSIDED;
            ld.sidenum[1] = UniqueSidedef(&wl->sides[1]);
        }
        else
            ld.sidenum[1] = -1;
        maplinedefstore->addElement(&ld);
    }
}

void ProcessThings(void)
{
    mapthing_t  mt;
    thing_t*    wt;

    mapthingstore = new Storage<mapthing_t>;

    for (wt = firstthing.next ; wt ; wt = wt->next)
    {
        memset (&mt,0,sizeof(mt));
        mt.x = (short)wt->x;
        mt.y = (short)wt->y;
        mt.angle =wt->angle;
        mt.type =wt->type;
        mt.options =wt->options;
        mapthingstore->addElement(&mt);
    }
}

void SaveBlocks(void);
extern int  numbsplines;
extern int  numsubsectors;
extern char wadfilename[];
extern int  gamemap;

void SaveDoomMap(void)
{
    wadinfo_t   wad;
    int         count;
    char        lumpname[9];

    numbsplines = 0;
    numsubsectors = 0;

    infostore = new Storage<lumpinfo_t>;

    fp = fopen(wadfilename, "wb");
    fwrite(&wad, sizeof(wadinfo_t), 1, fp); // leave space for wad header

    // write a label for the map name at the start of the wadfile
    if (gamemap<10)
        sprintf (lumpname,"map0%i", gamemap);
    else
        sprintf (lumpname,"map%i", gamemap);
    addName(lumpname, NULL, 0);

    // DWDt@C`ł́AZN^[͌ʉĂȂ̂ŁAŏLINEDEF(SIDEDEF)eZN^[𓯈ꎋĒoA
    // ŁATuZN^[ƂɏꏊႤZN^[𕪗Ă
    //BuildSectordefs ();
    ProcessThings ();
    ProcessLineSideDefs ();
    ProcessNodes ();
    ProcessSectors ();
    ProcessConnections ();

    OutputThings ();
    OutputLineDefs ();
    OutputSideDefs ();
    OutputVertexes ();
    OutputSegs ();
    OutputSubsectors ();
    OutputNodes ();
    OutputSectors ();
    OutputConnections ();

    SaveBlocks ();

    delete mapthingstore;
    delete mapsectorstore;
    delete mapvertexstore;
    delete mapsidedefstore;
    delete maplinedefstore;
    delete mapsubsecstore;
    delete mapsegstore;
    delete mapnodestore;
    count = subsecportals->count();
    for (int i=0 ; i<count ; i++)
        delete *(subsecportals->elementAt(i));
    delete subsecportals;

    // write the directory
    wad.infotableofs = ftell(fp);
    count = infostore->count();
    fwrite(infostore->elementAt(0), sizeof(lumpinfo_t), count, fp);

    // write the header
    strncpy(wad.identification, "PWAD", 4);
    wad.numlumps = count;
    fseek(fp, 0, SEEK_SET);
    fwrite(&wad, sizeof(wadinfo_t), 1, fp);

    fclose(fp);

    delete infostore;
}
