#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <stdio.h>
#include "thingdef.h"
#include "sysutils.h"

lumpinfo_t*     infotable;
int             numlumps;
char*           pnames;
int             numpatches;
void**          patches;
int*            texture1;
int             numtextures;
maptexture_t**  textures;
int             firstflat, lastflat;
int             numflats;
void**          flats;
char**          flatnames;
int             numthingdefs;
byte*           doompal;

int GetNumForName(char* name)
{
    int     length;
    int     i=0;

    length = strlen(name);
    while (i < numlumps)
    {
       if (!_strnicmp(infotable[i].name, name, length))
           break;
       i++;
    }
    return i;
}

void* CacheLumpNum(int i, FILE* fp)
{
    void* ptr;

    if (infotable[i].size == 0)
        return NULL;
    ptr = malloc(infotable[i].size);
    fseek(fp, infotable[i].filepos, SEEK_SET);
    fread(ptr, 1, infotable[i].size, fp);
    return ptr;
}

int ReadLump(char* name, FILE* fp, void** ptr)
{
    int	i;

    i = GetNumForName(name);
    if (i == numlumps)
        return -1;

    *ptr = CacheLumpNum(i, fp);
    return i;
}

BOOL LoadWadFile(char* filename)
{
    FILE*       fp;
    wadinfo_t   header;
    int         lumpnum;
    char*       name_p;
    char        name[9];
    int         i, j;
    byte*       base;
    int*        offset;
    void*       flat;

    fp = fopen(filename, "rb");
    if (!fp)
        return FALSE;

    fread(&header, sizeof(wadinfo_t), 1, fp);
    if (strncmp(header.identification, "IWAD", 4))
        return FALSE;
    numlumps = header.numlumps;

    infotable = (lumpinfo_t*)malloc(sizeof(lumpinfo_t)*numlumps);
    fseek(fp, header.infotableofs, SEEK_SET);
    fread(infotable, sizeof(lumpinfo_t), numlumps, fp);

    lumpnum = ReadLump("PNAMES", fp, (void**)&pnames);
    if (lumpnum == -1)
        return FALSE;
    numpatches = *((int *)pnames);
    patches = (void**)malloc(numpatches*sizeof(*patches));
    name_p = pnames+4;
    name[8] = 0;
    for (i=0 ; i<numpatches ; i++)
    {
        strncpy(name, name_p+i*8, 8);
        lumpnum = ReadLump(name, fp, &patches[i]);
        if (lumpnum == -1)
        {
            Debug("%s lump ܂\n", name);
            patches[i] = NULL;
        }
    }

    lumpnum = ReadLump("TEXTURE1", fp, (void**)&texture1);
    if (lumpnum == -1)
        return FALSE;
    numtextures = *texture1;
    textures = (maptexture_t**)malloc(numtextures*sizeof(*textures));
    base = (byte *)texture1;
    offset = texture1 + 1;
    for (i=0 ; i<numtextures ; i++, offset++)
        textures[i] = (maptexture_t *)(base + *offset);

    firstflat = GetNumForName("F_START") + 1;
    lastflat = GetNumForName("F_END") - 1;
    numflats = lastflat - firstflat + 1;
    flats = (void**)malloc(numflats*sizeof(*flats));
    flatnames = (char**)malloc(numflats*sizeof(*flatnames));
    for (i=0, j=0 ; i<numflats ; i++)
    {
        flat = CacheLumpNum(firstflat+i, fp);
        if (!flat)
            continue;
        flats[j] = flat;
        flatnames[j] = infotable[firstflat+i].name;
        j++;
    }
    numflats = j;

    numthingdefs = GetNumThingdefs();
    for (i=0 ; i<numthingdefs ; i++)
    {
        if (thingdefs[i].spritename)
        {
            lumpnum = ReadLump(thingdefs[i].spritename, fp, (void**)&thingdefs[i].spritedata);
            if (lumpnum == -1)
            {
                Debug("%s lump ܂\n", thingdefs[i].spritename);
                thingdefs[i].spritedata = NULL;
            }
        }
    }

    lumpnum = ReadLump("PLAYPAL", fp, (void**)&doompal);
    if (lumpnum == -1)
        return FALSE;

    fclose(fp);

    return TRUE;
}

void DrawColumnInBitmap(post_t* post, byte* pBits, int x, int originy, int pitch, int height) 
{
    int     count;
    int     toprow;
    byte*   source;
    int     i;
    byte*   dest;

    while (post->topdelta != 0xff)
    {
        count = post->length;
        toprow = originy + post->topdelta;
        source = (byte *)post + 3;

        if (toprow < 0)
        {
            count += toprow;
            source -= toprow; 
            toprow = 0;
        }

        if ((toprow + count) > height)
            count = height - toprow;

        if (count > 0)
        {
            for (i=0 ; i<count ; i++)
            {
                dest = pBits + x + (toprow + i)*pitch;
                *dest = source[i];
            }
        }

        post = (post_t *)((byte *)post + post->length + 4); // post
    }
}

void PatchToBitmap(patch_t* patch, PBITMAPINFO lpbmi, void** ppBits)
{
    int     pitch;
    int     i;
    post_t* column;

    pitch = (patch->width + 3) & ~3;
    *ppBits = calloc(pitch*patch->height, 1);

    for (i=0 ; i<patch->width ; i++)
    {
        column = (post_t *)((byte *)patch + patch->columnofs[i]);
        DrawColumnInBitmap(column, (byte*)*ppBits, i, 0, pitch, patch->height);
    }

    lpbmi->bmiHeader.biWidth = patch->width;
    lpbmi->bmiHeader.biHeight = -(patch->height);
}

void TextureToBitmap(maptexture_t* texture, PBITMAPINFO lpbmi, void** ppBits)
{
    int             pitch;
    int             i;
    patchdesc_t*    patchdesc;
    patch_t*        realpatch;
    int             x1, x2, x;
    post_t*         column;

    pitch = (texture->width + 3) & ~3;
    *ppBits = (void*)calloc(pitch*texture->height, 1);

    for (i=0, patchdesc=texture->patches ; i<texture->patchcount ; i++, patchdesc++)
    {
        realpatch = (patch_t*)patches[patchdesc->patch];
        x1 = patchdesc->originx;
        x2 = x1 + realpatch->width;

        if (x1 < 0)
            x = 0;
        else
            x = x1;

        if (x2 > texture->width)
            x2 = texture->width;

        for ( ; x<x2 ; x++)
        {
            column = (post_t *)((byte *)realpatch + realpatch->columnofs[x-x1]);
            DrawColumnInBitmap(column, (byte*)*ppBits, x, patchdesc->originy, pitch, texture->height);
        }
    }

    lpbmi->bmiHeader.biWidth = texture->width;
    lpbmi->bmiHeader.biHeight = -(texture->height);
}

int GetTextureNumForName(char* name)
{
    int     length;
    int     i;
    char    buf[9];

    buf[8] = 0;
    memcpy(buf, name, 8);
    length = strlen(buf);
    for (i=0 ; i<numtextures ; i++)
    {
        if (!(_strnicmp(textures[i]->name, buf, length)))
            return i;
    }

    return -1;  // '-'̎
}

char dash[] = "-";

char* GetTextureNameForNum(int i)
{
    if (i == -1)
        return dash;

    return textures[i]->name;
}

int GetFlatNumForName(char* name)
{
    int     length;
    int     i;
    char    buf[9];

    buf[8] = 0;
    memcpy(buf, name, 8);
    length = strlen(buf);
    for (i=0 ; i<numflats ; i++)
    {
        if (!(_strnicmp(flatnames[i], buf, length)))
            return i;
    }

    return -1;
}